@@ -1,13 +1,72 @@
+
+# V[ \nalba f(a), f(X)]
+# X - data locs
+# a - place where gradient is evaluated.
+cov_gen_dy <- function(a, X, l, sigma) {
+    K_yXya <- hetGP::cov_gen(X, matrix(a, nrow = 1), l, type = 'Gaussian')
+    return(sigma * t(as.numeric(K_yXya) * t(2*(t(X) - a) / l)))
+}
+
+#xloc <- X[1,]
+#grad(function(a) k(a, xloc), a)
+#cov_gen_dy(a, X, l, sigma)
+
+# V[\nabla f(a), \nabla f(a)]
+cov_gen_dd <- function(a, X, l, sigma) {
+    return(sigma*diag(2/l))
+}
+
+#jacobian(function(a2) grad(function(a1) k(a1, a2), a), a)
+#cov_gen_dd(a, X, l, sigma)
+#
+
+# Double check nugget, especially for Vdf_df term.
+C_at_a <- function(a, X, y, l, sigma, gpe, beta = 0, Ki = NULL) {
+    # Generate prior variances.
+    Vdf_df <- cov_gen_dd(a, X, l, sigma) 
+    Vdf_f <- cov_gen_dy(a, X, l, sigma)
+
+    if (missing(Ki)) {
+        Vf_f <- sigma*hetGP::cov_gen(X, X, l, type = 'Gaussian')
+        sol <- t(solve(Vf_f + diag(gpe, ncol = nrow(X), nrow = nrow(X)), t(Vdf_f)))
+    } else {
+        sol <- Vdf_f %*% Ki / sigma
+    }
+
+    # Get posterior moments via NCE. 
+    post_mean <- sol %*% (y - beta)
+    post_var <- Vdf_df - sol %*% t(Vdf_f)
+
+    #return(post_var + tcrossprod(post_mean))
+    return(post_var + tcrossprod(post_mean))
+}
+
+C_GP_empirical <- function(fit) {
+    Cxs <- lapply(1:nrow(fit$X), function(n) C_at_a(fit$X[n,], fit$X, fit$Z, fit$theta, fit$nu_hat, Ki = fit$Ki))
+    Cxh <- Reduce(function(x, y) x+y, Cxs) / nrow(fit$X)
+}
+
+C_GP_discrete <- function(fit, S) {
+    Cxs <- lapply(1:nrow(S), function(n) C_at_a(S[n,], fit$X, fit$Z, fit$theta, fit$nu_hat, Ki = fit$Ki))
+    Cxh <- Reduce(function(x, y) x+y, Cxs) / nrow(fit$X)
+}
+
 #' @title C matrix closed form expression for a GP.
 #' @description 
-#' Computes the integral (uniform measure) over the [0,1] box domain of the outer product of the gradients of a Gaussian process. 
-#' The corresponding matrix is the C matrix central in the active subspace methodology.
-#' @param model \code{homGP} or \code{hetGP} GP model, see \code{\link[hetGP]{hetGP-package}} containing, e.g.,
+#' Computes the integral over the input domain of the outer product of the gradients of a Gaussian process. 
+#' The corresponding matrix is the C matrix central in active subspace methodology.
+#' @param modelX This may be either 1) a \code{homGP} or \code{hetGP} GP model, see \code{\link[hetGP]{hetGP-package}} containing, e.g.,
 #'  a vector of \code{theta}s, type of covariance \code{ct}, an inverse covariance matrix \code{Ki},
-#' a design matrix \code{X0}, and response vector \code{Z0}.
+#' a design matrix \code{X0}, and response vector \code{Z0}. 2) A matrix of design locations, in which case a vector of responses must be given as the y argument, and this function will fit a default model for you.  
+#' @param y A vector of responses corresponding to the design matrix; may be ommited if a GP fit is provided in the modelX argument. 
+#' @param measure One of c("lebesgue", "gaussian", "trunc_gaussian", "sample", "discrete"), indiciating the probability distribution with respect to which the input points are drawn in the definition of the active subspace. "lebesgue" uses the Lebesgue or Uniform measure over the unit hypercube [0,1]^d. "gaussian" uses a Gaussian or Normal distribution, in which case xm and xv should be specified. "trunc_gaussian" gives a truncated Gaussian or Normal distribution over the unit hypercube [0,1]^d, in which case xm and xv should be specified. "sample" gives the Sample or Empirical measure (dirac deltas located at each design point), which is equivalent to calculating the average expected gradient outer product at the design points. "discrete" gives a measure which puts equal weight at points in the input space specified via the S parameter, which should be a matrix with one row for each atom of the measure.  
+#' @param xm If measure is "gaussian" or "trunc_gaussian", gives the mean vector. 
+#' @param xv If measure is "gaussian" or "trunc_gaussian", gives the marginal variance vector. The covariance matrix is assumed to be diagonal.
+#' @param S If measure is "discrete", gives the locations of the measure's atoms. 
+#' @param verbose Should we print progress?
 #' @return a \code{const_C} object with elements
 #' \itemize{
-#' \item \code{model}: GP model provided;
+#' \item \code{model}: GP model provided or estimated;
 #' \item \code{mat}: C matrix estimated;
 #' \item \code{Wij}: list of W matrices, of size number of variables;
 #' \item \code{ct}: covariance type (1 for "Gaussian", 2 for "Matern3_2", 3 for "Matern5_2").
@@ -18,13 +77,14 @@
 #' P. Constantine (2015), Active Subspaces, Philadelphia, PA: SIAM.
 #' @export
 #' @useDynLib activegp
-#' @importFrom  Rcpp evalCpp
+#' @importFrom  Rcpp evalCpp 
 #' @import hetGP
 #' @seealso \code{\link[activegp]{print.const_C}}, \code{\link[activegp]{plot.const_C}}
 #' @examples 
 #' ################################################################################
 #' ### Active subspace of a Gaussian process
 #' ################################################################################
+#' \donttest{ 
 #' library(hetGP); library(lhs)
 #' set.seed(42)
 #' 
@@ -60,67 +120,124 @@
 #' plot(C_hat)
 #' # par(mfrow = c(1, 1)) # restore graphical window
 #' 
-C_GP <- function(model){
-  # Extract model design and response 
-  design <- model$X0
-  response <- model$Z0
-  
-  # isotropic case:
-  if(length(model$theta) < ncol(model$X0)) model$theta <- rep(model$theta, ncol(model$X0))
-  
-  # Discover covariance type
-  if (model$covtype == "Gaussian") {
-    M_num <- 1
-    ct <- 1
-    theta <- sqrt(model$theta/2)
-  } else if (model$covtype == "Matern3_2") {
-    M_num <- 3
-    ct <- 2
-    theta <- model$theta
-  } else if (model$covtype == "Matern5_2") {
-    M_num <- 5/3
-    ct <- 3
-    theta <- model$theta
-  } else {
-    stop("Unsupported Covariance type in model passed to C_GP.")
-  }
-  
-  nvar <- ncol(design)
-  if (nrow(design) != length(response)) {
-    stop("microbenchmark")
-  }
-  if(max(design) > 1 + sqrt(.Machine$double.eps)|| min(design) < 0 - sqrt(.Machine$double.eps)) {
-      warning("Designs are supposed to be in [0,1]^d; please rescale.\n Extreme values of ", min(design), ", ", max(design), " detected in design. ")
-  }
-  
-  # Create a const_C object to return
-  C <- list()
-  C$mat <- matrix(0, nvar, nvar)
-  class(C) <- "const_C"
-  C$Wij <- lapply(1:nvar, function(i) list())
-  C$model <- model
-  # C$response <- response
-  # C$design <- design
-  C$ct <- ct
-  
-  # Populate its elements
-  # Ki <- C$model$Ki 
-  Kir <- crossprod(model$Ki, response)
-  for(i in 1:nvar)
-    for(j in i:nvar){
-      Wij <- W_kappa_ij(design = design, theta = theta, i1 = i - 1, i2 = j - 1, ct = ct)
-      
-      M <- M_num/theta[i]^2 * (i == j) - sum(model$Ki * Wij) + crossprod(Kir, Wij %*% Kir) 
-      C$mat[i,j] <- C$mat[j,i] <- M
-      C$Wij[[i]][[j]] <- Wij
+#' } 
+C_GP <- function(modelX, y, measure = 'lebesgue', xm = NULL, xv = NULL, S = NULL, verbose = TRUE) {
+    if ('matrix' %in% class(modelX)) {
+        if (missing(y)) {
+            stop("If no model is provided, both X and y must be.")
+        }
+        X <- modelX
+        if (verbose) cat("Fitting GP...\n")
+        model <- hetGP::mleHomGP(X, y)
+    } else if ('homGP' %in% class(modelX) || 'list' %in% class(modelX)) {
+        model <- modelX
+    } else {
+        stop("Unrecognized first argument to C_GP: Should be either a design matrix X (to be accompanied by a response vector y) or a GP fit, typically the result of a call to hetGP::mleHomGP.")
     }
-  return(C)
+    if (verbose) cat("Calculating Active Subspace Matrix...\n")
+
+    int_measures <- c("lebesgue", "gaussian", "trunc_gaussian")
+    sum_measures <-  c("sample", "discrete")
+    allowed_measures <- c(int_measures, sum_measures)
+    measure <- match.arg(measure, allowed_measures, several.ok = FALSE)
+    if (is.null(xm)) xm <- 0
+    if (is.null(xv)) xv <- 1
+
+    if (measure %in% c('lebesgue', 'trunc_gaussian'))
+      if(max(model$X) > 1 + sqrt(.Machine$double.eps)|| min(model$X) < 0 - sqrt(.Machine$double.eps)) {
+          warning("Designs are supposed to be in [0,1]^d for lebesgue or trunc_gaussian; please rescale.\n Extreme values of ", min(model$X), ", ", max(model$X), " detected in model$X. ")
+      }
+
+    if (is.null(model$X)) {
+        stop("Model object needs to have design points in X")
+    }
+    # isotropic measure:
+    if (length(xm) == 1) {
+        xm = rep(xm,ncol(model$X))
+    } else if (length(xm) != ncol(model$X)) {
+        stop("Please either supply a scalar or a vector of length equal to the dimension of the data for the Gaussian mean parameter.")
+    }
+    if (length(xv) == 1) {
+        xv = rep(xv,ncol(model$X))
+    } else if (length(xv) != ncol(model$X)) {
+        stop("Please either supply a scalar or a vector of length equal to the dimension of the data for the Gaussian variance parameter.")
+    }
+    if (length(model$theta) == 1) {
+        model$theta <- rep(model$theta,ncol(model$X))
+    } 
+
+    Kir <- model$Ki %*% model$Z0
+    # Create a const_C object to return
+    C <- list()
+
+    if (model$covtype == 'Gaussian') {
+        theta <- sqrt(model$theta/2)
+        ct <- 1
+    } else if (model$covtype == 'Matern3_2') {
+        theta <- model$theta
+        ct <- 2
+    }else if (model$covtype == 'Matern5_2') {
+        theta <- model$theta
+        ct <- 3
+    } else {
+        stop("Unsupported Covariance type in model passed to C_GP")
+    }
+
+    if (measure %in% int_measures) {
+        measure_ind <- which(measure==int_measures)-1
+        #C$mat <- quick_C(measure_ind, model$X, model$Ki, Kir, theta, xm = xm, xv = xv, ct = ct)
+        ret <- quick_C(measure_ind, model$X, model$Ki, Kir, theta, xm = xm, xv = xv, ct = ct, verbose)
+        C$mat <- ret$C
+        C$Wij <- ret$W
+    } else if (measure %in% sum_measures) {
+        #stop("Not implemented.")
+        if (measure == 'sample') {
+            S <- model$X
+        } else if (measure == 'discrete'){
+            if (is.null(S)) {
+                stop("For the discrete measure, you must specify the atoms as rows in the matrix S.")
+            }
+        } else {
+            stop("Measure not recognized. ")
+        }
+        C$mat <- C_GP_discrete(model, S) 
+    }
+
+    class(C) <- "const_C"
+    #TODO: See if we can back these out.
+    #C$Wij <- lapply(1:ncol(model$X), function(i) list())
+    C$model <- model
+    C$ct <- which(c('Gaussian', 'Matern3_2', 'Matern5_2')==model$covtype)#TODO: Generalize kernels
+    C$measure <- measure
+
+    return(C)
 }
 
-if(!isGeneric("update")) {
-  setGeneric(name = "update",
-             def = function(object, ...) standardGeneric("update")
-  )
+#' Active Subspace Prewarping
+#'
+#' Computes a matrix square root of C = Lt %*% t(Lt).
+#' @param ... Parameters to be passed to C_GP, if C was not provided. 
+#' @param C the result of a call to C_GP. If provided, all other arguments are ignored. 
+#' @return The matrix Lt which can be used for sensitivity prewarping, i.e. by computing Xw = X %*% Lt.
+#' @export
+Lt_GP <- function(..., C) {
+    if (missing(C)) {
+        C <- C_GP(...) 
+    }
+    ed <- eigen(C)
+    Lt <- ed$vectors %*% diag(sqrt(ed$values))
+    return(Lt)
+}
+
+#' Extract Matrix
+#'
+#' Given a const_C object, extracts the actual matrix itself.
+#' @param x A const_C object with field 'mat'.
+#' @param ... Additional parameters. Not used. 
+#' @return The mat entry of C, a matrix.
+#' @export
+as.matrix.const_C <- function(x, ...) {
+    return(x$mat)
 }
 
 #' C update with new observations
@@ -131,8 +248,8 @@ if(!isGeneric("update")) {
 #' @param Znew vector of size \code{nrow(Xnew)} for the new responses at \code{Xnew}
 #' @param ... not used (for consistency of update method)
 #' @return The updated const_C object originally provided. 
-#' @seealso \code{\link[activegp]{C_GP}} to generate const_C objects from \code{\link[hetGP]{mleHomGP}} objects; \code{\link[activegp]{update_C2}} for an update using faster expressions.  
 #' @export
+#' @seealso \code{\link[activegp]{C_GP}} to generate const_C objects from \code{\link[hetGP]{mleHomGP}} objects; \code{\link[activegp]{update_C2}} for an update using faster expressions.  
 #' @useDynLib activegp
 #' @importFrom  Rcpp evalCpp
 #' @examples
@@ -180,10 +297,10 @@ update.const_C <- function(object, Xnew, Znew, ...){
   if(is.null(nrow(Xnew))) Xnew <- matrix(Xnew, nrow = 1)
   n0 <- nrow(object$model$X0) # to identify replicates
   nvar <- ncol(Xnew)
-  
+
   # Update ancillary quantities.
   object$model <- update(object$model, Xnew = Xnew, Znew = Znew, maxit = 0)
-  
+
   theta <- object$model$theta
   # isotropic case:
   if(length(theta) < ncol(object$model$X0)) theta <- rep(theta, nvar)
@@ -193,10 +310,10 @@ update.const_C <- function(object, Xnew, Znew, ...){
   }else{
     if(object$ct == 2) M_num <- 3 else M_num <- 5/3
   } 
-  
+
   n <- nrow(object$model$X0)
   Kir <- crossprod(object$model$Ki, object$model$Z0)
-  
+
   for(i in 1:nvar) {
     for(j in i:nvar) {
       # If xnew is a replicate, n = n0 and Wijs are unchanged
@@ -207,14 +324,14 @@ update.const_C <- function(object, Xnew, Znew, ...){
       }else{
         Wij <- object$Wij[[i]][[j]]
       }
-      
+
       # Cov(dYi(X),dYj(X))
       M <- M_num/theta[i]^2 * (i == j) - sum(object$model$Ki * Wij) + crossprod(Kir, Wij) %*% Kir 
       object$mat[i,j] <- object$mat[j,i] <- M
     }
   }
   return(object)
-  
+
 }
 
 #' Update Constantine's C, using update formula
@@ -270,6 +387,7 @@ update_C2 <- function(C, xnew, ynew){
   return(Cup)
 }
 
+
 #' Expected variance of trace of C 
 #' 
 #' @param C A const_C object, the result of a call to \code{\link[activegp]{C_GP}}.
@@ -430,6 +548,7 @@ C_var <- function(C, xnew, grad = FALSE){
 #' ################################################################################
 #' ### Norm of the variance of C criterion landscape
 #' ################################################################################
+#' \donttest{ 
 #' library(hetGP)
 #' set.seed(42)
 #' nvar <- 2
@@ -461,6 +580,7 @@ C_var <- function(C, xnew, grad = FALSE){
 #' Cvar_grid <- apply(Xgrid, 1, cvar_crit, C = C_hat)
 #' filled.contour(matrix(Cvar_grid, ngrid), color.palette = terrain.colors,
 #'                plot.axes = {axis(1); axis(2); points(design, pch = 20)})
+#' }
 C_var2 <- function(C, xnew, grad = FALSE){
   if(is.null(nrow(xnew))) xnew <- matrix(xnew, nrow = 1)
   nvar <- ncol(xnew)
@@ -490,6 +610,7 @@ C_var2 <- function(C, xnew, grad = FALSE){
 #' and the third is the same vector as would have been returned if grad was \code{FALSE}: simply the values of beta and gamma.
 #' @keywords internal
 get_betagamma <- function(C, xnew, grad = FALSE) {
+  if (C$measure != 'lebesgue') stop("Sequential design only supports Lebesgue measure currently.")
   if(is.null(nrow(xnew))) xnew <- matrix(xnew, nrow = 1)
   nvar <- ncol(xnew)
   kn1 <- cov_gen(xnew, C$model$X0, theta = C$model$theta, type = C$model$covtype)
@@ -531,7 +652,8 @@ get_betagamma <- function(C, xnew, grad = FALSE) {
     for (j in i:nvar) {
       wa <- drop(W_kappa_ij2(C$model$X0, xnew, theta = theta, i - 1, j - 1, ct = C$ct))  # w(X, xnew)
       wb <- drop(W_kappa_ij2(xnew, C$model$X0, theta = theta, i - 1, j - 1, ct = C$ct))  # c(w(xnew, X), w(xnew, xnew))
-      Wij <- C$Wij[[i]][[j]]
+      #Wij <- C$Wij[[i]][[j]]
+      Wij <- W_kappa_ij(design = C$model$X0, theta = theta, i1 = i - 1, i2 = j - 1, ct = C$ct)
       kntKiWij <- crossprod(Kikn, Wij)
       WijKiKn <- Wij %*% Kikn
       
@@ -586,8 +708,8 @@ get_betagamma <- function(C, xnew, grad = FALSE) {
 #' @param ... Additional parameters. Not used. 
 #' @export
 print.const_C <- function(x, ...) {
-  cts <- c("Gaussian", "Matern3_2", "Matern5_2")
-  cat(paste(cts[x$ct], "kernel GP Estimate of Constantine's C:\n"))
+  cts <- c("Gaussian", "Matern3_2", "Matern5_2", 'Discrete')
+  cat(paste(cts[x$ct], "kernel GP Estimate of Constantine's C wrt",x$measure,"measure:\n"))
   print(x$mat)
 }
 
@@ -602,7 +724,7 @@ plot.const_C <- function(x, output = c("all", "matrix", "logvals", "projfn"), ..
   output <- match.arg(output)
   if(output %in% c("all", "matrix")) image(x$mat, main = "C matrix values heatmap")
   if(output %in% c("all", "logvals")) plot(log(eigen(x$mat)$values), main = "log eigen values of C", xlab = "index", ylab = "")
-  if(output %in% c("all", "projfn")) plot(x$model$X0 %*% eigen(x$mat)$vectors[,1], x$model$Y0, xlab = "First AS direction", ylab = "Function values")
+  if(output %in% c("all", "projfn")) plot(x$model$X0 %*% eigen(x$mat)$vectors[,1], x$model$Z0, xlab = "First AS direction", ylab = "Function values")
 
 }
 
